{-# LANGUAGE DeriveDataTypeable,TypeSynonymInstances,FlexibleInstances #-}
module Unification (unify,match,unifyList,subsumes,subsumesCheck,subst,substList,fv,fn,vList) where

import Types
import qualified Data.Map as Map
import qualified Data.List as List
import qualified Data.Maybe as Maybe
import Control.Monad.State
import Control.Exception

class Unifiable a where
    unify :: (Maybe Subst) -> a -> a -> (Maybe Subst)

instance Unifiable Term where
    unify (Just rho) (Var x) (Var y) =
        if x == y then (Just rho)
        else case Map.lookup (Var x) rho of
               Just t -> unify (Just rho) (Var y) t
               Nothing -> Just $ Map.insert (Var x) (Var y) rho
    unify (Just rho) (Var x) (Func ty f fArgs) =
        case Map.lookup (Var x) rho of
          Just t -> unify (Just rho) t (Func ty f fArgs)
          Nothing -> if occurs (Var x) (Func ty f fArgs) then Nothing
                     else Just $ Map.insert (Var x) (Func ty f fArgs) rho
    unify (Just rho) (Func ty f fArgs) (Var x) =
        unify (Just rho) (Var x) (Func ty f fArgs)
    unify (Just rho) (Func fTy f fArgs) (Func gTy g gArgs) =
        if f == g && fTy == gTy then unifyList (Just rho) fArgs gArgs
        else Nothing
    unify Nothing _ _ = Nothing

match :: (Maybe Subst) -> Pattern -> Term -> (Maybe Subst)
match (Just rho) (Bind x _) t = Just $ Map.insert (Var x) t rho
match (Just rho) (Match t1) t2 = unify (Just rho) t1 (subst rho t2)
match (Just rho) (Format f1 args1) (Func TFunc f2 args2) =
  if f1 /= f2 then Nothing
  else matchList (Just rho) args1 (substList rho args2)
  where
    matchList (Just rho) [] [] = Just rho
    matchList (Just rho) (x:xs) (y:ys) = matchList (match (Just rho) x (subst rho y)) xs ys
    matchList Nothing _ _ = Nothing
match (Just rho) (Format f1 args1) (Var x) = match (Just rho) (Format f1 args1) (subst rho (Var x))
match Nothing _ _ = Nothing

occurs :: Term -> Term -> Bool
occurs (Var x) (Var y) = x == y
occurs (Var x) (Func ty f fArgs) = List.any (occurs (Var x)) fArgs
occurs t1 t2 = error $ "Occurs takes a variable as first argument, got " ++ show t1 ++ " and " ++ show t2 ++ " instead."

unifyList :: Unifiable a => Maybe Subst -> [a] -> [a] -> Maybe Subst
unifyList (Just rho) (x:xs) (y:ys) =
    unifyList (unify (Just rho) x y) xs ys
unifyList (Just rho) [] [] = Just rho
unifyList (Just rho) _ _ = Nothing
unifyList Nothing _ _ = Nothing

class Substitutable a where
    subst :: Subst -> a -> a

instance Substitutable Term where
    subst sigma (Var x) =
        case Map.lookup (Var x) sigma of
          Nothing -> (Var x)
          Just t -> if Var x == t then t else subst sigma t
    subst sigma (Func ty f args) =
        case Map.lookup (Func ty f args) sigma of
          Just t -> if Func ty f args == t then t else subst sigma t
          Nothing -> Func ty f (substList sigma args)

instance Substitutable Pattern where
  subst sigma (Bind name t) = bindTerm (subst sigma (Var name)) (subst sigma t)
    where
      bindTerm (Var x) t = Bind x t
      bindTerm (Func TFunc f1 args1) (Func TFunc f2 args2) =
        if f1 == f2 then Format f1 $ List.map (uncurry bindTerm) (List.zip args1 args2) else error "Type mismatch"
      bindTerm (Func TName f1 args1) (Func TName f2 []) = (Match $ Func TName f1 args1)
  subst sigma (Match t) = Match (subst sigma t)
  subst sigma (Format name args) = Format name (List.map (subst sigma) args)

instance Substitutable Pred where
    subst sigma (Pred p args) = Pred p (substList sigma args)
    subst sigma (Neq t1 t2) = Neq (subst sigma t1) (subst sigma t2)

instance Substitutable Clause where
    subst sigma (Clause h c) = Clause (substList sigma h) (subst sigma c)

instance Substitutable Process where
    subst sigma Nil = Nil
    subst sigma (Par p1 p2 pos) = Par (subst sigma p1) (subst sigma p2) pos
    subst sigma (Repl p pos) = Repl (subst sigma p) pos
    subst sigma (New pa p pos) = New (subst sigma' pa) (subst sigma' p) pos
       where sigma' = restrict sigma pa
    subst sigma (In ch pa p pos) = In (subst sigma ch) (subst sigma' pa) (subst sigma' p) pos
       where sigma' = restrict sigma pa
    subst sigma (Out ch msg p pos) = Out (subst sigma ch) (subst sigma msg) (subst sigma p) pos
    subst sigma (Let pa f p1 p2 pos) = Let (subst sigma' pa) (subst sigma f) (subst sigma' p1) (subst sigma' p2) pos
       where sigma' = restrict sigma pa
    subst sigma (IfE b p1 p2 pos) = IfE (subst sigma b) (subst sigma p1) (subst sigma p2) pos
    subst sigma (SetE b p pos) = SetE (subst sigma b) (subst sigma p) pos
    subst sigma (SelectE pa b p pos) = SelectE (subst sigma' pa) (subst sigma' b) (subst sigma' p) pos
       where sigma' = restrict sigma pa
    subst sigma (Event s x ty p pos) = Event (substVar sigma s) (subst sigma x) ty (subst sigma p) pos
    subst sigma (Lock s p pos) = Lock (substVars sigma s) (subst sigma p) pos
    subst sigma (Unlock s p pos) = Unlock (substVars sigma s) (subst sigma p) pos

instance Substitutable SetExp where
    subst sigma (SetIn t s) = SetIn (subst sigma t) (substVar sigma s)
    subst sigma (And e1 e2) = And (subst sigma e1) (subst sigma e2)
    subst sigma (Or e1 e2) = Or (subst sigma e1) (subst sigma e2)
    subst sigma (Not e) = Not (subst sigma e)

instance (Substitutable a) => Substitutable (Maybe a) where
    subst sigma (Just x) = Just (subst sigma x)
    subst sigma (Nothing) = Nothing

substVar sigma x =
  case Map.lookup (Var x) sigma of
    Just (Var y) -> y
    Just _ -> error $ "expecting variable to variable mapping from " ++ x ++ " in " ++ show sigma
    Nothing -> x

substVars sigma = List.map (substVar sigma)

substList :: Substitutable a => Subst -> [a] -> [a]
substList sigma = List.map (subst sigma)

restrict sigma pa = List.foldl (flip Map.delete) sigma (bv pa)

class Subsumable a where
    subsumes :: a -> a -> Bool
    subsumesCheck :: a -> a -> Maybe Subst -> Maybe Subst

instance Subsumable Term where
    subsumes x y = case subsumesCheck x y (Just Map.empty) of
                     Just rho -> True
                     Nothing -> False
    subsumesCheck (Var name1) y rho = unify rho (Var name1) y
    subsumesCheck (Func tf1 f1 args1) (Func tf2 f2 args2) rho =
        if tf1 == tf2 && f1 == f2 then subsumesList args1 args2 rho else Nothing
    subsumesCheck (Func tf1 f1 args1) y rho = Nothing

subsumesList (x:xs) (y:ys) rho = subsumesList xs ys rho'
    where rho' = subsumesCheck x y rho
subsumesList [] [] rho = rho
subsumesList xs ys rho =
    error $ "Unifying lists " ++ show xs ++ " and " ++ show ys ++ " with different lengths"

instance Unifiable Pred where
    unify rho (Pred p1 pargs1) (Pred p2 pargs2) =
        if p1 == p2 then
            List.foldl foldUnify rho $ List.zip pargs1 pargs2
        else Nothing
        where foldUnify rho (x,y) = unify rho x y

instance Subsumable Pred where
    subsumes p1 p2 = case subsumesCheck p1 p2 (Just Map.empty) of
                       Just rho -> True
                       Nothing -> False

    subsumesCheck (Pred p1 pargs1) (Pred p2 pargs2) rho =
        if p1 == p2 then subsumesList pargs1 pargs2 rho else Nothing

instance Subsumable Clause where
    subsumes (Clause h1 c1) (Clause h2 c2) =
        if List.length h1 <= List.length h2 then
            not $ List.null $ Maybe.catMaybes $ findAll h1 h2 [rho]
        else False
        where rho = subsumesCheck c1 c2 (Just Map.empty)
              findOne p1 (p2:h2) rho =
                  case subsumesCheck p2 p1 rho of
                    Just rho' -> Just rho' : findOne p1 h2 rho
                    Nothing -> findOne p1 h2 rho
              findOne p1 [] rho = []
              findAll (p1:h1) h2 rhos =
                  case h1 of
                    [] -> rhos'
                    _ -> findAll h1 h2 rhos'
                  where rhos' = List.concatMap (findOne p1 h2) rhos
              findAll [] h2 rhos = rhos
    subsumesCheck _ _ _ = error "subsumesCheck unapplicable to clauses"

class FV a where
  fv :: a -> [Term]
class FN a where
  fn :: a -> [Term]
class BV a where
  bv :: a -> [Term]

vList f l = List.nub $ List.concatMap f l

instance FV Term where
  fv (Var x) = [Var x]
  fv (Func TName _ _ ) = []
  fv (Func TFunc f args) = vList fv args

instance FN Term where
  fn (Var x) = []
  fn (Func TName name nargs) = [Func TName name nargs]
  fn (Func TFunc f fargs) = vList fn fargs

instance FV Pattern where
  fv (Bind x t) = fv t
  fv (Match t) = fv t
  fv (Format name args) = vList fv args

instance BV Pattern where
  bv (Bind x t) = [Var x]
  bv (Match t) = []
  bv (Format name args) = vList bv args

instance FV Pred where
  fv (Pred n args) = vList fv args

instance FN Pred where
  fn (Pred n args) = vList fn args

instance FV Process where
  fv (Nil) = []
  fv (Par p1 p2 pos) = List.nub $ fv p1 ++ fv p2
  fv (Repl p pos) = fv p
  fv (New pa p pos) = (List.nub $ fv p ++ fv pa) List.\\ (bv pa)
  fv (In ch pa p pos) = (List.nub $ fv ch ++ fv pa ++ fv p) List.\\ (bv pa)
  fv (Out ch msg p pos) = List.nub $ fv ch ++ fv msg ++ fv p
  fv (Let t1 t2 p1 p2 pos) = (List.nub $ fv t1 ++ fv t2 ++ fv p1 ++ fv p2) List.\\ (bv t1)
  fv (IfE b p1 p2 pos) = List.nub $ fv b ++ fv p1 ++ fv p2
  fv (SetE b p pos) = List.nub $ fv b ++ fv p
  fv (SelectE pa b p pos) = (List.nub $ (case b of Just b -> fv b; Nothing -> []) ++ fv pa ++ fv p) List.\\ bv pa
  fv (Event s x ty p pos) = List.nub $ fv x ++ fv p
  fv (Lock s p pos) = fv p
  fv (Unlock s p pos) = fv p

instance FV SetExp where
    fv (SetIn t s) = fv t
    fv (And e1 e2) = List.nub $ fv e1 ++ fv e2
    fv (Or e1 e2) = List.nub $ fv e1 ++ fv e2
    fv (Not e) = fv e

instance FN SetExp where
    fn (SetIn t s) = fn t
    fn (And e1 e2) = List.nub $ fn e1 ++ fn e2
    fn (Or e1 e2) = List.nub $ fn e1 ++ fn e2
    fn (Not e) = fn e
